from collections.abc import Callable

import tensorflow as tf
from keras import activations
from keras.saving import register_keras_serializable
from keras.src.layers.merging.base_merge import _Merge
from tensorflow.python.ops.nn_ops import leaky_relu, relu6

from ..layers.base import HLayerBase
from ..utils import apf_to_tuple, tuple_to_apf


@register_keras_serializable(package="HGQ")
class HQuantize(HLayerBase):
    def __init__(self, paq_conf=None, beta=0., **kwargs):
        super().__init__(
            paq_conf=paq_conf,
            beta=beta,
            **kwargs
        )

    @tf.function(jit_compile=True)
    def forward(self, x, training=None, record_minmax=None):
        return self.paq(x, training=training, record_minmax=record_minmax)  # type: ignore

    def compute_output_shape(self, input_shape):
        return input_shape


@register_keras_serializable(package="HGQ")
class HActivation(HLayerBase, tf.keras.layers.Activation):
    def __init__(self, activation, beta=0., paq_conf=None, **kwargs):
        super().__init__(activation=activation, beta=beta, paq_conf=paq_conf, **kwargs)

    def post_build(self, input_shape):
        if self.paq_config['rnd_strategy'] == 'auto':
            self.paq_config['rnd_strategy'] = 'standard_round'

        if self.paq_config['rnd_strategy'] not in ('floor', 3):
            # Jit when using standard round. Very unlikely to have errors.
            self.forward: Callable = tf.function(jit_compile=True)(self.__forward)  # type: ignore

        elif self.activation in (activations.relu, activations.linear, relu6, leaky_relu):
            # tf 213 scatters everthing around. Terrible design choice.
            # Jit when activation is more or less linear.
            self.forward: Callable = tf.function(jit_compile=True)(self.__forward)  # type: ignore

        else:
            self.forward: Callable = self.__forward

        super().post_build(input_shape)

    def __forward(self, x, training=None, record_minmax=None):
        x = self.activation(x)  # type: ignore
        return self.paq(x, training=training, record_minmax=record_minmax)  # type: ignore

    def compute_output_shape(self, input_shape):
        return input_shape


@register_keras_serializable(package="HGQ")
class HAdd(HLayerBase, _Merge):

    @tf.function(jit_compile=True)
    def forward(self, inputs, training=None, record_minmax=None):
        output = tf.reduce_sum(inputs, axis=0)
        return self.paq(output, training=training, record_minmax=record_minmax)  # type: ignore

    def compute_output_shape(self, input_shape):
        return input_shape[0]


# class HMultply(HLayerBase, _Merge):

#     @tf.function(jit_compile=True)
#     def forward(self, inputs, training=None, record_minmax=None):
#         output = inputs[0]
#         for i in range(1, len(inputs)):
#             output *= inputs[i]
#         return self.paq(output, training=training, record_minmax=record_minmax)  # type: ignore

#     def compute_output_shape(self, input_shape):
#         return input_shape[0]



import os
from collections.abc import Callable

import keras
import numpy as np
import tensorflow as tf
from keras.layers import Layer

from HGQ.proxy.fixed_point_quantizer import gfixed_quantizer
from HGQ.proxy.precision_derivation import get_input_kifs, get_produced_kif, get_result_kifRS
from HGQ.utils import apf_to_tuple, tuple_to_apf

LUT_SIZE_LIMITATION = int(os.environ.get('LUT_SIZE_LIMITATION', 2**12))


@keras.utils.register_keras_serializable(package='HGQ')
class UnaryLUT(Layer):
    proxy_ready = True

    def __init__(self, kif_in: tuple[int, int, int], kif_out: tuple[int, int, int], RND='TRN', SAT='WRAP', **kwargs):
        assert sum(kif_in) > 0, 'Input to activation is constantly zero'
        assert sum(kif_out) > 0, 'Output of activation is constantly zero'
        if LUT_SIZE_LIMITATION > 0:
            assert 2**sum(kif_in) < LUT_SIZE_LIMITATION, f'Input to activation is too large ({2**sum(kif_in)} > {LUT_SIZE_LIMITATION}). If you want to raise this limit, set the LUT_SIZE_LIMITATION environment variable.'
        self.kif_in = kif_in
        self.kif_out = kif_out
        k, i, f = kif_in
        self.scale = 2. ** f
        self.table = None
        if (table := kwargs.pop('table', None)) is not None:
            k, i, f, = kif_out
            k, b, i = k, k + i + f, k + i
            table = gfixed_quantizer(table, k, b, i, RND, SAT)  # type:ignore
            self.table = tf.Variable(table, dtype='float32', trainable=False, name='table')
        super().__init__(**kwargs)

    def call(self, inputs, **kwargs):  # type:ignore
        if not self.built:
            self.build(inputs.shape)
        inputs = tf.round(inputs * self.scale)
        inputs = inputs % self.table.shape[0]  # type:ignore
        return tf.gather(self.table, tf.cast(inputs, 'int32'))

    def build(self, input_shape):
        super().build(input_shape)
        if self.table is not None:
            return
        N = 2**sum(self.kif_in)
        self.table = tf.Variable(tf.zeros(N), dtype='float32', trainable=False, name='table')

    @classmethod
    def from_activation(cls, activation: Layer | Callable, kif_in=None, kifRS_out=None):

        if kif_in is None:
            kifs_in = get_input_kifs(activation)
            assert len(kifs_in) == 1, f'Activation function {activation} has more than one input. Please specify the input dtype.'
            kif_in = kifs_in[0]

        kifRS_out = kifRS_out or get_result_kifRS(activation)
        kif_out = kifRS_out[:3]
        R, S = kifRS_out[-2:]

        k, i, f = kif_in
        kif_in = k, i, f
        assert k + i + f > 0, 'Activation function is applied to an zero array. Something is wrong.'
        N = 2**(k + i + f)
        assert N < int(os.environ.get('HLS_MAX_ACTIVATION_LUT_SIZE', 2**16)), f'Input to activation function is too large ({N} > {os.environ.get("HLS_MAX_ACTIVATION_LUT_SIZE", 2**16)}). If you want to raise this limit, set the HLS_MAX_ACTIVATION_LUT_SIZE environment variable.'
        if k:
            inp_table = np.empty(N, dtype=np.float64)
            inp_table[:N // 2] = np.linspace(0, 2.**i - 2.**-f, N // 2, dtype=np.float64)
            inp_table[N // 2:] = inp_table[:N // 2] - 2.**i
        else:
            inp_table = np.linspace(-2.**i * k, 2.**i - 2.**-f, N, dtype=np.float64)
        table: np.ndarray = np.array(activation(inp_table), dtype=np.float32)
        return cls(kif_in, kif_out, table=table, RND=R, SAT=S)

    def get_config(self):
        config = super().get_config()
        config.update({
            'kif_in': self.kif_in,
            'kif_out': self.kif_out,
        })
        return config


def xfr_to_unary_lut(layer: keras.layers.Layer, max_table_size=1024):
    if not isinstance(layer, keras.layers.Activation):
        return layer
    if layer.activation is keras.activations.softmax:
        return layer  # simply doesn't work
    if layer.activation in (keras.activations.relu, keras.activations.linear):
        return layer  # not necessary
    kifs_in = get_input_kifs(layer)
    if len(kifs_in) > 1:
        return layer
    if 2**sum(*kifs_in) > max_table_size:
        return layer
    kif_in = kifs_in[0]

    return UnaryLUT.from_activation(layer, kif_in=kif_in)


get_produced_kif.register(UnaryLUT, lambda x: x.kif_out)



import pytest

try:
    import qkeras
    from qkeras import quantizers
except ImportError:
    pytest.skip('qkeras not installed', allow_module_level=True)

import numpy as np
from helpers import get_test_dir, run_model_test, set_seed
from tensorflow import keras


def create_model():

    l = lambda X: qkeras.QDense(X, kernel_quantizer='quantized_bits(8, 3, alpha=1.)', activation='quantized_bits(5, 3, alpha=1.)', bias_quantizer='quantized_bits(8, 3, alpha=1)')
    c = lambda X: qkeras.QConv2D(X, 2, kernel_quantizer=quantizers.quantized_bits(8, 3, alpha=1.), activation=quantizers.quantized_bits(6, 1, alpha=1., symmetric=True), bias_quantizer=quantizers.quantized_bits(8, 3, alpha=1))
    r = lambda X: keras.layers.Activation('relu')(X)
    rr = lambda X: qkeras.QActivation('quantized_relu(5, 3)')(X)

    L = l(10)

    inp = keras.layers.Input((10,))
    q_inp = qkeras.QActivation(qkeras.quantized_bits(8, 4, alpha=1.))(inp)
    out = L(q_inp)
    out1 = L(out)
    out1 = r(out1)
    out1 = L(out1)
    out2 = l(20)(out1)
    out2 = r(out2)
    out3 = l(20)(out2)
    out2 = keras.layers.Reshape((2, 5, 2))(out2)
    out2 = c(2)(out2)
    out2 = rr(out2)
    out2 = keras.layers.Flatten()(out2)
    out = keras.layers.Concatenate()([out2, out3])
    model = keras.models.Model(inp, out)
    return model


def get_data(N: int, sigma: float, max_scale: float, seed):
    rng = np.random.default_rng(seed)
    a1 = rng.normal(0, sigma, (N, 10)).astype(np.float32)
    a2 = rng.uniform(0, max_scale, (1, 10)).astype(np.float32)
    return (a1 * a2).astype(np.float32)


@pytest.mark.parametrize("N", [50000, 10])
@pytest.mark.parametrize("io_type", ['io_parallel', 'io_stream'])
@pytest.mark.parametrize("aggressive", [False])
@pytest.mark.parametrize("backend", ['vivado', 'vitis', 'Quartus'])
@pytest.mark.parametrize("seed", [42, 114514, 1919810])
def test_syn_qkeras(N: int, io_type: str, aggressive: bool, backend: str, seed: int):
    dir = get_test_dir()
    set_seed(seed)
    model = create_model()
    data = get_data(N, 8, 1, seed)

    run_model_test(model, None, data, io_type, backend, dir, aggressive, skip_sl_test=True)


# if __name__ == '__main__':
#     test_syn_small(10, 'auto', 'io_parallel', 0.49, True, 'vivado', 1919810)



from collections.abc import Callable

import numpy as np
import tensorflow as tf
from keras.activations import linear, relu

from ..quantizer import HGQ
from ..utils import apf_to_tuple, get_default_kq_conf, get_default_paq_conf, tuple_to_apf, warn


@staticmethod  # type: ignore
@tf.function(jit_compile=True)
def scale_grad(x, scale):
    sx = x * scale
    return sx + tf.stop_gradient(x - sx)


class ABSBaseLayer(tf.keras.layers.Layer):

    @property
    def last_layer(self):
        assert len(self._inbound_nodes) == 1, f'input_container is only available for layers used only once. {self.name} is used {len(self._inbound_nodes)} times.'
        assert not isinstance(self._inbound_nodes[0].inbound_layers, list), f'input_container is only available for layers with a single input. {self.name} has {len(self._inbound_nodes[0].inbound_layers)} inputs.'
        return self._inbound_nodes[0].inbound_layers

    @property
    def input_bw(self):
        assert len(self._inbound_nodes) <= 1, f"Layer {self.name} is reused {len(self._inbound_nodes)} times. This is not allowed."
        try:
            return self.last_layer.act_bw
        except AssertionError:
            return None

    @property
    def input_bw_exact(self):
        assert len(self._inbound_nodes) <= 1, f"Layer {self.name} is reused {len(self._inbound_nodes)} times. This is not allowed."
        return self.last_layer.act_bw_exact.astype(np.float32)


class HLayerBase(ABSBaseLayer):
    """Abstract base class for all layers in the library. Child classes should call post_build() after calling their build() method.
    """

    def __init__(self, kq_conf=None, paq_conf=None, beta=0., **kwargs):
        self._has_kernel = None
        self._has_bias = None
        self.kq_config = kq_conf or get_default_kq_conf()
        "kernel quantizer config"
        self.paq_config = paq_conf or get_default_paq_conf()
        "pre-activation quantizer config"
        self.beta = tf.Variable(beta, dtype=tf.float32, name='beta', trainable=False)
        "BOPs-regularization strength"
        self.record_minmax = False
        self._has_last_layer = False
        self._do_adapt_kernel_bits = kwargs.pop('do_adapt_kernel_bits', True)
        self._delayed_kernel_bits_adaption = False

        activation = kwargs.get('activation', None)
        if activation is not None and type(self).__name__ != 'HActivation':
            if isinstance(activation, str):
                activation = activation.lower()
            elif isinstance(activation, Callable):
                activation = activation.__name__.lower()
            assert activation in ['relu', 'linear'], f'activation other than relu and linear are only supported for HActivation layer, but is defined for {type(self).__name__} layer.'

        super().__init__(**kwargs)

    def build(self, input_shape):
        super().build(input_shape)
        self.post_build(input_shape)

    @property
    def can_bias_cover_rnd(self):
        if not self._has_bias:
            return False
        quantizer_shape: tuple[int, ...] = tuple(self.paq.fbw.shape)
        bias_shape: tuple[int, ...] = tuple(self.bias.shape)
        if len(bias_shape) != 1:
            warn(f'bias shape {bias_shape} is not supported.')
            return False
        if np.prod(quantizer_shape) == 1:
            return True
        if self.channel_loc == -1:
            return np.prod(quantizer_shape[:-1]) == 1 and bias_shape == quantizer_shape[-1:]
        elif self.channel_loc == 1:
            return np.prod(quantizer_shape[1:]) == 1 and bias_shape == quantizer_shape[0:1]
        return False

    @property
    def _relu_act(self):
        return hasattr(self, 'activation') and self.activation is tf.keras.activations.relu

    def post_build(self, input_shape):
        """This method should be called after calling build() method of the child class. It initializes the quantizers and sets the bops variable, and set a few flags (_has_kernel, _has_bias, _relu_act) for convenience.)"""
        self._has_kernel = False
        self._has_bias = False
        if hasattr(self, 'kernel') and self.kernel is not None:
            self._has_kernel = True
        if hasattr(self, 'bias') and self.bias is not None:
            self._has_bias = True

        self.init_quantizers(input_shape)
        if not hasattr(self, 'channel_loc'):
            self.channel_loc = -1
        if self.paq_config['rnd_strategy'] == 'auto':
            self.paq.rnd_strategy = 0 if self.can_bias_cover_rnd else 3

        self.bops = tf.Variable(0, dtype=tf.float32, trainable=False, name='bops')

    def init_quantizers(self, input_shape):
        """Initializes the High Granularity Quantizers for the kernel and the pre-activation values. This method is called by post_build() method."""

        if self._has_kernel:
            kq = HGQ.from_config(self.kq_config)
            bw_shape, degeneracy = kq._compute_bw_shape_and_degeneracy(self.kernel.shape)
            fbw = self.add_weight(
                name='kernel_bw',
                shape=bw_shape,
                initializer=tf.keras.initializers.Constant(kq.init_bw),  # type: ignore
                trainable=kq.trainable and self.trainable,
                regularizer=kq.regularizer
            )
            kq.build(fbw)
            kq.degeneracy = degeneracy
            if self._do_adapt_kernel_bits and not self._delayed_kernel_bits_adaption:
                kq.adapt_bw_bits(self.kernel)
                self._do_adapt_kernel_bits = False
            self.kq = kq

        aq = HGQ.from_config(self.paq_config)
        output_shape = self.compute_output_shape(input_shape)
        bw_shape, degeneracy = aq._compute_bw_shape_and_degeneracy(output_shape)
        fbw = self.add_weight(
            name='activation_bw',
            shape=bw_shape,
            initializer=tf.keras.initializers.Constant(aq.init_bw),  # type: ignore
            trainable=aq.trainable and self.trainable,
            regularizer=aq.regularizer
        )

        aq.build(fbw)
        aq.degeneracy = degeneracy
        self.paq = aq

    def call(self, x, training=None, record_minmax=None):
        if not self.built:
            self.build(tuple(x.shape))

        if record_minmax is None:
            record_minmax = training or self.record_minmax

        dtype = self.dtype or tf.keras.backend.floatx()
        x = tf.cast(x, dtype)

        return self.forward(x, training=training, record_minmax=record_minmax)

    def forward(self, x, training=None, record_minmax=None):
        raise NotImplementedError

    @property
    @tf.function(jit_compile=True)
    def kernel_bw(self):
        """Returns (over) approximated bitwidth of the kernel. Differentiable."""
        int_bits, fp_bits, kn = self.kq.get_bits(self.fused_kernel)  # type: ignore
        k_bw = tf.nn.relu(int_bits + fp_bits)  # negative sign not considered for kernel
        k_bw = scale_grad(k_bw, tf.sqrt(1. / self.kq.degeneracy))  # type: ignore
        return tf.broadcast_to(k_bw, self.kernel.shape)

    @tf.function(jit_compile=True)
    def _kernel_bw(self, qk):
        """Returns (over) approximated bitwidth of the kernel. Differentiable. Takes the differentiable quantized kernel as input to avoid recomputing."""
        int_bits, fp_bits, kn = self.kq.get_bits(qk, quantized=True)  # type: ignore
        k_bw = tf.nn.relu(int_bits + fp_bits)  # negative sign not considered for kernel
        k_bw = scale_grad(k_bw, tf.sqrt(1. / self.kq.degeneracy))  # type: ignore
        return tf.broadcast_to(k_bw, qk.shape)

    @property
    def kernel_bw_exact(self):
        """Returns exact bitwidth of the kernel. Non-differentiable. For post-training use."""
        kn, int_bits, fb = self.kq.get_bits_exact(self.fused_kernel)
        return int_bits + fb  # sign not considered for kernel

    @property
    def act_bw(self):
        """Returns the bitwidth of the pre-activation values. Differentiable."""
        int_bits, fp_bits, kn = self.paq.get_bits(pos_only=self._relu_act)  # type: ignore
        bw = int_bits + fp_bits
        if not self._relu_act:
            bw = bw + kn
        bw = tf.nn.relu(bw)
        bw = scale_grad(bw, tf.sqrt(1. / self.paq.degeneracy))  # type: ignore
        return tf.broadcast_to(bw, (1,) + self.output_shape[1:])

    @property
    def act_bw_exact(self) -> np.ndarray:
        """Returns the exact bitwidth of the pre-activation values. Non-differentiable. For post-training use."""
        kn, int_bits, fb = self.paq.get_bits_exact(pos_only=self._relu_act)
        bw = int_bits + fb + kn
        return np.broadcast_to(bw, (1,) + self.output_shape[1:])

    @property
    def fused_bias(self):
        return self.bias

    @property
    def fused_kernel(self):
        return self.kernel

    @property
    @tf.function(jit_compile=True)
    def fused_qkernel(self):
        """Returns the final, quantized kernel for deployment. non-differentiable, should not be used for training."""
        return self.kq(self.fused_kernel, training=False)  # type: ignore

    @property
    @tf.function(jit_compile=True)
    def fused_qbias(self):
        """Returns the final, quantized bias for deployment. non-differentiable, should not be used for training. When using rounding to nearest and the bias can cover the rounding error, bias is pre-biased to cover the rounding shift 2^-fbw, and then TRN can be used instead RND without any loss of accuracy."""
        bias = self.paq.bias_forward(self.fused_bias, False, self.channel_loc)  # type: ignore

        fbw = self.paq.fbw
        if self.channel_loc == -1:
            dims = list(range(len(fbw.shape) - 1))
        elif self.channel_loc == 1:
            dims = [0] + list(range(2, len(fbw.shape)))
        else:
            raise ValueError('channel_loc must be -1 or 1')

        fbw = tf.reduce_max(self.paq.fbw, axis=dims, keepdims=False)
        fbw = tf.broadcast_to(fbw, bias.shape)
        mask = tf.reduce_max(self.act_bw, axis=dims, keepdims=False) > 0

        if self.paq.rnd_strategy != 3 and self.can_bias_cover_rnd:
            bias = tf.pow(2., -tf.round(fbw) - 1) + bias  # type: ignore

        return tf.where(mask, bias, 0.)

    def reset_minmax(self):
        """Resets the recorded minmax values for the pre-activation quantizer."""
        self.paq.minmax_reg_reset()  # type: ignore

    @property
    def compute_exact_bops(self):
        """Computes the exact bops for the layer. Non-differentiable. For post-training use."""
        self.bops.assign(tf.constant(0.))
        return np.float32(0.)

    def get_config(self):
        base_config = super().get_config()
        config = dict(
            kq_conf=self.kq_config,
            paq_conf=self.paq_config,
            beta=float(self.beta.numpy())  # type: ignore
        )
        return {**base_config, **config}

    def get_keras_config(self):
        config = super().get_config()
        return config
